from __future__ import annotations

import logging
from collections.abc import Iterator, Mapping
from enum import Enum, auto
from pathlib import Path
from typing import Any, TypeVar

import tomlkit
from pydantic import BaseModel, Field, PrivateAttr, field_validator
from pydantic_settings import (
    BaseSettings,
    PydanticBaseSettingsSource,
    SettingsConfigDict,
    TomlConfigSettingsSource,
)

from esgpull.constants import CONFIG_FILENAME
from esgpull.exceptions import BadConfigError, VirtualConfigError
from esgpull.install_config import InstallConfig
from esgpull.models.options import Option, Options

logger = logging.getLogger("esgpull")
T = TypeVar("T")


def _get_root() -> Path:
    if InstallConfig.current is not None:
        return InstallConfig.current.path
    else:
        return InstallConfig.default


class Paths(BaseModel, validate_assignment=True, validate_default=True):
    data: Path = Path("data")
    db: Path = Path("db")
    log: Path = Path("log")
    tmp: Path = Path("tmp")
    plugins: Path = Path("plugins")

    @field_validator(
        "data",
        "db",
        "log",
        "tmp",
        "plugins",
        mode="after",
    )
    @classmethod
    def _set_path_from_root(cls, value: Path) -> Path:
        root = _get_root()
        if not value.is_absolute():
            value = root / value
        return value

    def values(self) -> Iterator[Path]:
        yield self.data
        yield self.db
        yield self.log
        yield self.tmp
        yield self.plugins


class Credentials(BaseModel, validate_assignment=True):
    filename: str = "credentials.toml"


class Cli(BaseModel, validate_assignment=True):
    page_size: int = 20


class Db(BaseModel, validate_assignment=True):
    filename: str = "esgpull.db"


class Download(BaseModel, validate_assignment=True):
    chunk_size: int = 1 << 26
    http_timeout: int = 20
    max_concurrent: int = 5
    disable_ssl: bool = False
    disable_checksum: bool = False
    show_filename: bool = False


class DefaultOptions(BaseModel, validate_assignment=True):
    distrib: str = Options._distrib_.name
    latest: str = Options._latest_.name
    replica: str = Options._replica_.name
    retracted: str = Options._retracted_.name

    @field_validator(
        "distrib", "latest", "replica", "retracted", mode="before"
    )
    @classmethod
    def _is_valid_option(cls, value: str | Option) -> str:
        if isinstance(value, str):
            return Option(value.lower()).name
        else:
            return value.name

    def asdict(self) -> dict[str, str]:
        return dict(
            distrib=self.distrib,
            latest=self.latest,
            replica=self.replica,
            retracted=self.retracted,
        )


class API(BaseModel, validate_assignment=True):
    index_node: str = "esgf-node.ipsl.upmc.fr"
    http_timeout: int = 20
    max_concurrent: int = 5
    page_limit: int = 50
    default_options: DefaultOptions = Field(default_factory=DefaultOptions)
    default_query_id: str = ""
    use_custom_distribution_algorithm: bool = False


class Plugins(BaseModel, validate_assignment=True):
    enabled: bool = False


class ConfigKind(Enum):
    Virtual = auto()
    NoFile = auto()
    Partial = auto()
    Complete = auto()


class ConfigKey:
    path: tuple[str, ...]

    def __init__(
        self,
        first: str | tuple[str, ...] | list[str],
        *rest: str,
    ) -> None:
        if isinstance(first, (tuple, list)):
            self.path = tuple(first) + rest
        elif "." in first:
            self.path = tuple(first.split(".")) + rest
        else:
            self.path = (first,) + rest

    def __iter__(self) -> Iterator[str]:
        yield from self.path

    def __hash__(self) -> int:
        return hash(str(self.path))

    def __eq__(self, other: object) -> bool:
        match other:
            case ConfigKey():
                return self.path == other.path
            case _:
                raise TypeError(type(other))

    def __repr__(self) -> str:
        return ".".join(self)

    def __add__(self, path: str) -> ConfigKey:
        return ConfigKey(self.path, path)

    def __len__(self) -> int:
        return len(self.path)

    def exists_in(self, source: Mapping | None) -> bool:
        if source is None:
            return False
        doc = source
        for key in self:
            if key in doc:
                doc = doc[key]
            else:
                return False
        return True

    def value_of(self, source: Any) -> Any:
        doc = source
        for key in self:
            try:
                doc = doc[key]
            except TypeError:
                doc = getattr(doc, key)
        return doc


def fix_rename_search_api(doc: dict) -> dict:
    if "api" in doc and "search" in doc:
        raise KeyError(
            "Both 'api' and 'search' (deprecated) are used in your "
            "config, please use 'api' only."
        )
    elif "search" in doc:
        logger.warn(
            "Deprecated key 'search' is used in your config, "
            "please use 'api' instead."
        )
        doc["api"] = doc.pop("search")
    return doc


def fix_remove_auth(doc: dict) -> dict:
    if "paths" in doc and "auth" in doc["paths"]:
        logger.warn(
            "Deprecated 'paths.auth' is present in your config, "
            "you can remove it safely."
        )
        doc["paths"].pop("auth")
    return doc


def iter_keys(
    source: Mapping,
    path: ConfigKey | None = None,
) -> Iterator[ConfigKey]:
    for key in source.keys():
        if path is None:
            local_path = ConfigKey(key)
        else:
            local_path = path + key
        if isinstance(source[key], Mapping):
            yield from iter_keys(source[key], local_path)
        else:
            yield local_path


def pop_and_clear_empty_parents(source: Mapping, ckey: ConfigKey):
    *parent_path, last_key = ckey.path
    parent_ckey = ConfigKey(parent_path)
    parent_ckey.value_of(source).pop(last_key)

    for i in range(len(parent_path), 0, -1):
        parent_ckey = ConfigKey(parent_path[: i - 1])
        container_ckey = ConfigKey(parent_path[:i])
        parent = parent_ckey.value_of(source)
        container = container_ckey.value_of(source)
        if isinstance(container, dict) and len(container) == 0:
            parent.pop(container_ckey.path[-1])
        else:
            break  # Stop if we hit a non-empty container


config_fixers = [fix_rename_search_api, fix_remove_auth]


class TomlKitConfigSettingsSource(TomlConfigSettingsSource):
    def _read_file(self, file_path: Path) -> dict[str, Any]:
        with open(file_path, mode="rb") as toml_file:
            doc = tomlkit.load(toml_file)
            for fixer in config_fixers:
                try:
                    doc = fixer(doc)
                except Exception:
                    raise BadConfigError(file_path)
            return dict(doc)


class Config(BaseSettings):
    # TODO: set in a load method instead
    # model_config = SettingsConfigDict(toml_file=_get_root() / "config.toml")
    model_config = SettingsConfigDict(toml_file=None)

    paths: Paths = Field(default_factory=Paths)
    credentials: Credentials = Field(default_factory=Credentials)
    cli: Cli = Field(default_factory=Cli)
    db: Db = Field(default_factory=Db)
    download: Download = Field(default_factory=Download)
    api: API = Field(default_factory=API)
    plugins: Plugins = Field(default_factory=Plugins)
    _raw: dict[str, Any] | None = PrivateAttr(default=None)
    _config_file: Path | None = PrivateAttr(default=None)

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> tuple[PydanticBaseSettingsSource, ...]:
        return (
            init_settings,
            TomlKitConfigSettingsSource(settings_cls),
        )

    @classmethod
    def load(cls, path: Path) -> Config:
        try:
            file_path = path / CONFIG_FILENAME
            cls.model_config["toml_file"] = file_path
            instance = cls()
        finally:
            cls.model_config["toml_file"] = None
        instance._config_file = file_path
        if file_path.is_file():
            with file_path.open("rb") as toml_file:
                instance._raw = tomlkit.load(toml_file)
        else:
            instance._raw = None
        return instance

    @classmethod
    def default(cls) -> Config:
        ## TODO: rename+deprecate
        ## very bad name since this is loading from the **default** config
        ## file path, as set in InstallConfig
        return cls.load(path=_get_root())

    @property
    def kind(self) -> ConfigKind:
        if self._config_file is None:
            return ConfigKind.Virtual
        elif not self._config_file.is_file():
            return ConfigKind.NoFile
        elif self.unset_options():
            return ConfigKind.Partial
        else:
            return ConfigKind.Complete

    def dump(self, with_defaults: bool = True) -> dict:
        result = self.model_dump(mode="json")
        if not with_defaults:
            unset = set(self.unset_options())
            for ckey in iter_keys(self.model_dump()):
                if ckey in unset:
                    pop_and_clear_empty_parents(result, ckey)
        return result

    def unset_options(self) -> list[ConfigKey]:
        result: list[ConfigKey] = []
        dump = self.model_dump()
        for ckey in iter_keys(dump):
            if not ckey.exists_in(self._raw):
                result.append(ckey)
        return result

    def update_item(
        self,
        key: str,
        value: T,
        empty_ok: bool = False,
    ) -> T:
        if self._raw is None and empty_ok:
            self._raw = {}
        elif self._raw is None:
            raise VirtualConfigError
        doc = self._raw
        obj = self
        ckey = ConfigKey(key)
        *parts, last = ckey
        for part in parts:
            doc.setdefault(part, {})
            doc = doc[part]
            obj = getattr(obj, part)
        old_value = getattr(obj, last)
        setattr(obj, last, value)
        doc[last] = value
        return old_value

    def set_default(self, key: str) -> Any:
        ckey = ConfigKey(key)
        if self._raw is None:
            raise VirtualConfigError()
        elif not ckey.exists_in(self._raw):
            return None
        default_config = Config()
        default_value: Any = ckey.value_of(default_config)
        old_value: Any = ckey.value_of(self)

        *parent_path, last_key = ckey.path
        parent_ckey = ConfigKey(parent_path)
        obj = parent_ckey.value_of(self)
        setattr(obj, last_key, default_value)
        pop_and_clear_empty_parents(self._raw, ckey)
        return old_value

    def generate(self, overwrite: bool = False) -> None:
        match (self.kind, overwrite):
            case (ConfigKind.Virtual, _):
                raise VirtualConfigError
            case (ConfigKind.Partial, overwrite):
                defaults = self.model_dump()
                for ckey in self.unset_options():
                    self.update_item(str(ckey), ckey.value_of(defaults))
            case (ConfigKind.Partial | ConfigKind.Complete, _):
                raise FileExistsError(self._config_file)
            case (ConfigKind.NoFile, _):
                self._raw = self.model_dump(mode="json")
            case _:
                raise ValueError(self.kind)
        self.write()

    def write(self) -> None:
        if self._config_file is None or self._raw is None:
            raise VirtualConfigError
        self._config_file.parent.mkdir(parents=True, exist_ok=True)
        with self._config_file.open("w") as f:
            tomlkit.dump(self._raw, f)



"""query_add_column_updated_at

Revision ID: 14c72daea083
Revises: c7c8541fa741
Create Date: 2025-05-07 14:49:43.993125

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '14c72daea083'
down_revision = 'c7c8541fa741'
branch_labels = None
depends_on = None


def upgrade() -> None:
    with op.batch_alter_table('query', schema=None) as batch_op:
        batch_op.add_column(sa.Column('updated_at', sa.DateTime(), server_default=sa.text('(CURRENT_TIMESTAMP)'), nullable=True))

    # Backfill nulls
    op.execute('UPDATE query SET updated_at = CURRENT_TIMESTAMP WHERE updated_at IS NULL')

    # Make non-nullable
    with op.batch_alter_table('query', schema=None) as batch_op:
        batch_op.alter_column('updated_at', nullable=False)


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('query', schema=None) as batch_op:
        batch_op.drop_column('updated_at')

    # ### end Alembic commands ###



import click
import rich
from click_params import PUBLIC_URL

from esgpull.context import Context


@click.command()
@click.option("--metadata", "-m", is_flag=True, default=False)
@click.argument("url", nargs=1, type=PUBLIC_URL)
def get(
    url: str,
    metadata: bool,  # display file metadata, no download
):
    """
    Steps:
        1. validate url (version+filename)
        2. create Context
        3. fetch File from Context
        4. search database (if exist -> [Y/N])
        5. check file exists in output directory (if exist -> [Y/N]) -> rename
        6. download
    """
    rich.print(url)
    ctx = Context()
    print(ctx)

    raise NotImplementedError()



from pathlib import Path

from click.testing import CliRunner

from esgpull import Esgpull
from esgpull.cli.add import add
from esgpull.cli.config import config as config_cmd
from esgpull.cli.remove import remove
from esgpull.cli.update import update
from esgpull.config import Config


def test_update_after_remove(root: Path, config: Config):
    config.generate(overwrite=True)
    runner = CliRunner()
    result_config = runner.invoke(config_cmd, ["api.page_limit", "10000"])
    assert result_config.exit_code == 0
    result_add = runner.invoke(
        add,
        [
            "activity_id:CMIP",
            "experiment_id:1pctCO2",
            "source_id:CNRM-ESM2-1",
            "--distrib",
            "true",
            "--track",
        ],
    )
    assert result_add.exit_code == 0
    result_update = runner.invoke(update, ["--yes"])
    assert result_update.exit_code == 0

    esg = Esgpull(root)  # Reinitialize to ensure fresh data
    esg.graph.load_db()
    query_id = list(esg.graph._shas)[0]
    query = esg.graph.get(query_id)
    nb_files_after_update = len(query.files)
    assert nb_files_after_update > 0

    result_remove = runner.invoke(remove, [query_id, "--yes"])
    assert result_remove.exit_code == 0

    result_add = runner.invoke(
        add,
        [
            "activity_id:CMIP",
            "experiment_id:1pctCO2",
            "source_id:CNRM-ESM2-1",
            "--distrib",
            "true",
            "--track",
        ],
    )
    assert result_add.exit_code == 0
    result_update = runner.invoke(update, ["--yes"])
    assert result_update.exit_code == 0

    esg = Esgpull(root)  # Reinitialize to ensure fresh data
    query = esg.graph.get(query_id)
    assert nb_files_after_update == len(query.files)



from __future__ import annotations

import logging
from collections.abc import AsyncIterator
from dataclasses import dataclass
from datetime import datetime
from functools import cached_property, partial
from pathlib import Path
from warnings import warn

from rich.live import Live
from rich.progress import (
    BarColumn,
    DownloadColumn,
    MofNCompleteColumn,
    Progress,
    ProgressColumn,
    SpinnerColumn,
    TaskID,
    TextColumn,
    TimeRemainingColumn,
    TransferSpeedColumn,
)

from esgpull.config import Config
from esgpull.context import Context
from esgpull.database import Database
from esgpull.download import DownloadCtx
from esgpull.exceptions import (
    DownloadCancelled,
    InvalidInstallPath,
    NoInstallPath,
    UnknownDefaultQueryID,
)
from esgpull.fs import Filesystem
from esgpull.graph import Graph
from esgpull.install_config import InstallConfig
from esgpull.models import (
    Facet,
    File,
    FileStatus,
    LegacyQuery,
    Options,
    Query,
    sql,
)
from esgpull.models.utils import short_sha
from esgpull.plugin import (
    Event,
    PluginManager,
    emit,
    get_plugin_manager,
    set_plugin_manager,
)
from esgpull.processor import Processor
from esgpull.result import Err, Ok, Result
from esgpull.tui import UI, DummyLive, ErrorCountColumn, Verbosity, logger
from esgpull.utils import format_size


@dataclass(repr=False)
class Esgpull:
    path: Path
    config: Config
    ui: UI
    db: Database
    context: Context
    fs: Filesystem
    graph: Graph

    def __init__(
        self,
        path: Path | str | None = None,
        verbosity: Verbosity = Verbosity.Detail,
        install: bool = False,
        record: bool = False,
        safe: bool = False,
        load_db: bool = True,
    ) -> None:
        if path is not None:
            path = Path(path)
            InstallConfig.choose(path=path)
            default = path
            warning = f"Using unknown location: {path}\n"
        else:
            default = InstallConfig.default
            warning = f"Using default location: {default}\n"
        if InstallConfig.current is None:
            if safe:
                raise NoInstallPath
            InstallConfig.choose(path=default)
            if InstallConfig.current_idx is None:
                idx = InstallConfig.add(default)
                InstallConfig.choose(idx=idx)
                needs_install = True
            else:
                idx = InstallConfig.current_idx
                needs_install = False
            self.path = InstallConfig.installs[idx].path
            warning += "To disable this warning, please run:\n"
            if needs_install:
                warning += f"$ esgpull self install {self.path}"
            else:
                warning += f"$ esgpull self choose {self.path}"
            if logger.level == logging.NOTSET:
                warn(warning)
            else:
                logger.warning(warning)
        else:
            self.path = InstallConfig.current.path
        if not install and not self.path.is_dir():
            raise InvalidInstallPath(path=self.path)
        self.config = Config.load(path=self.path)
        Options._set_defaults(**self.config.api.default_options.asdict())
        self.fs = Filesystem.from_config(self.config, install=install)
        self.ui = UI.from_config(
            self.config,
            verbosity=verbosity,
            record=record,
        )
        self.context = Context(self.config, noraise=True)
        if load_db:
            self.db = Database.from_config(self.config)
            self.graph = Graph(self.db)
        # Initialize plugin system
        plugin_config_path = self.config.paths.plugins / "plugins.toml"
        try:
            self.plugin_manager = get_plugin_manager()
            self.plugin_manager.__init__(config_path=plugin_config_path)
        except ValueError:
            self.plugin_manager = PluginManager(config_path=plugin_config_path)
            set_plugin_manager(self.plugin_manager)
        if self.config.plugins.enabled:
            self.plugin_manager.enabled = True
            self.config.paths.plugins.mkdir(exist_ok=True, parents=True)
            self.plugin_manager.discover_plugins(self.config.paths.plugins)

    def fetch_index_nodes(self) -> list[str]:
        """
        Returns a list of ESGF index nodes.

        Fetch hints from ESGF search API with a distributed query.
        """

        default_index = self.config.api.index_node
        logger.info(f"Fetching index nodes from {default_index!r}")
        options = Options(distrib=True)
        query = Query(options=options)
        facets = ["index_node"]
        hints = self.context.hints(
            query,
            file=False,
            facets=facets,
            index_node=default_index,
        )
        return list(hints[0]["index_node"])

    def fetch_facets(self, update: bool = False) -> bool:
        """
        Fill db with all existing facets found in ESGF index nodes.

        1. Fetch index nodes from `Esgpull.fetch_index_nodes()`
        2. Fetch all facets (names + values) from all index nodes.

        Workaround method, since searching directly for all facets using
        `distrib=True` seems to crash the index node.
        """

        # those facets have (almost) unique values
        IGNORE_NAMES = [
            "version",
            # "cf_standard_name",
            # "variable_long_name",
            "creation_date",
            # "datetime_end",
        ]
        nb_facets = self.db.scalars(sql.count_table(Facet))[0]
        logger.info(f"Found {nb_facets} facets in database")
        if nb_facets and not update:
            return False
        index_nodes = self.fetch_index_nodes()
        options = Options(distrib=False)
        query = Query(options=options)
        hints_coros = []
        for index_node in index_nodes:
            hints_results = self.context.prepare_hints(
                query,
                file=False,
                facets=["*"],
                index_node=index_node,
            )
            hints_coros.append(self.context._hints(*hints_results))
        hints = self.context.sync_gather(*hints_coros)
        new_facets: set[Facet] = set()
        facets_db = self.db.scalars(sql.facet.all())
        for index_hints in hints:
            for name, values in index_hints[0].items():
                if name in IGNORE_NAMES:
                    continue
                for value in values.keys():
                    facet = Facet(name=name, value=value)
                    if facet not in facets_db:
                        facet.compute_sha()
                        new_facets.add(facet)
        self.db.add(*new_facets)
        return len(new_facets) > 0

    @cached_property
    def legacy_query(self) -> Query:
        legacy = LegacyQuery
        if (
            legacy_db := self.db.get(Query, "LEGACY")
        ) and legacy_db is not None:
            legacy = legacy_db
        # else:
        # self.db.add(legacy)
        # self.graph.add(legacy, clone=False)
        # self.graph.merge(commit=True)
        return legacy

    def import_synda(
        self,
        url: Path,
        track: bool = False,
        size: int = 5000,
        ask: bool = False,
    ) -> int:
        assert url.is_file()
        synda = Database(f"sqlite:///{url}", run_migrations=False)
        synda_ids = synda.scalars(sql.synda_file.ids())
        shas = set(self.db.scalars(sql.file.linked()))
        msg = f"Found {len(synda_ids)} files to import, proceed?"
        if ask and not self.ui.ask(msg):
            return 0
        synda_shas: set[str] = set()
        idx_range = range(0, len(synda_ids), size)
        if track:
            iter_idx_range = self.ui.track(idx_range)
        else:
            iter_idx_range = iter(idx_range)
        nb_imported = 0
        for start in iter_idx_range:
            stop = min(len(synda_ids), start + size)
            ids = synda_ids[start:stop]
            synda_files = synda.scalars(sql.synda_file.with_ids(*ids))
            files: list[File] = []
            for synda_file in synda_files:
                file = synda_file.to_file()
                if file.sha not in shas:
                    file.queries.append(self.legacy_query)
                    files.append(file)
                    synda_shas.add(file.sha)
            if files:
                nb_imported += len(files)
                self.db.add(*files)
        return nb_imported

    # def add(
    #     self,
    #     *queries: Query,
    #     with_file: bool = False,
    # ) -> tuple[list[Query], list[Query]]:
    #     """
    #     Add new queries to query/options/selection tables.
    #     Returns two lists: added and discarded queries
    #     """
    #     for query in
    #     self.graph.add()
    #     return [], []

    # def install(
    #     self,
    #     *files: File,
    #     status: FileStatus = FileStatus.Queued,
    # ) -> tuple[list[File], list[File]]:
    #     """
    #     Insert `files` with specified `status` into db if not already there.
    #     """
    #     file_ids = [f.file_id for f in files]
    #     with self.db.select(File.file_id) as stmt:
    #         stmt.where(File.file_id.in_(file_ids))
    #         existing_file_ids = set(stmt.scalars)
    #     to_install = [f for f in files if f.file_id not in existing_file_ids]
    #     to_download: list[File] = []
    #     already_on_disk: list[File] = []
    #     for file in to_install:
    #         if status == FileStatus.Done:
    #             # skip check on status=done
    #             file.status = status
    #             to_download.append(file)
    #             continue
    #         path = self.fs.path_of(file)
    #         if path.is_file():
    #             file.status = FileStatus.Done
    #             already_on_disk.append(file)
    #         else:
    #             file.status = status
    #             to_download.append(file)
    #     self.db.add(*to_install)
    #     return to_download, already_on_disk

    # def remove(self, *files: File) -> list[File]:
    #     """
    #     Remove `files` from db and delete from filesystem.
    #     """
    #     file_ids = [f.file_id for f in files]
    #     with self.db.select(File) as stmt:
    #         stmt.where(File.file_id.in_(file_ids))
    #         deleted = stmt.scalars
    #     for file in files:
    #         if file.status == FileStatus.Done:
    #             self.fs.delete(file)
    #     self.db.delete(*deleted)
    #     return deleted

    # def autoremove(self) -> list[File]:
    #     """
    #     Search duplicate files and keep latest version only.
    #     """
    #     deprecated = self.db.get_deprecated_files()
    #     return self.remove(*deprecated)

    async def iter_results(
        self,
        processor: Processor,
        progress: Progress,
        task_ids: dict[str, TaskID],
        live: Live | DummyLive,
    ) -> AsyncIterator[Result[DownloadCtx]]:
        async for result in processor.process():
            task_idx = progress.task_ids.index(task_ids[result.data.file.sha])
            task = progress.tasks[task_idx]
            progress.update(task.id, visible=True)
            match result:
                case Ok():
                    progress.update(task.id, completed=result.data.completed)
                    if task.finished:
                        # TODO: add checksum verif here
                        progress.stop_task(task.id)
                        progress.update(task.id, visible=False)
                        sha = f"[b blue]{task.fields['sha']}[/]"
                        file = result.data.file
                        digest = result.data.digest
                        match self.fs.finalize(file, digest=digest):
                            case Ok():
                                size = f"[green]{format_size(int(task.completed))}[/]"
                                if task.elapsed is not None:
                                    final_speed = int(
                                        task.completed / task.elapsed
                                    )
                                    speed = (
                                        f"[red]{format_size(final_speed)}/s[/]"
                                    )
                                else:
                                    speed = "[b red]?[/]"
                                data_node = (
                                    f"[blue]{task.fields['data_node']}[/]"
                                )
                                parts = [sha, size, speed, data_node]
                                if self.config.download.show_filename:
                                    parts.append(task.fields["filename"])
                                msg = " · ".join(parts)
                                logger.info(msg)
                                live.console.print(msg)
                                yield result
                            case Err(_, err):
                                progress.remove_task(task.id)
                                yield Err(result.data, err=err)
                case Err():
                    progress.remove_task(task.id)
                    yield result
                case _:
                    raise ValueError("Unexpected result")

    async def download(
        self,
        queue: list[File],
        use_db: bool = True,
        show_progress: bool = True,
    ) -> tuple[list[File], list[Err]]:
        """
        Download files provided in `queue`.
        """
        for file in queue:
            file.status = FileStatus.Starting
        main_progress = self.ui.make_progress(
            SpinnerColumn(),
            MofNCompleteColumn(),
            TimeRemainingColumn(compact=True, elapsed_when_finished=True),
            ErrorCountColumn(),
        )
        file_columns: list[str | ProgressColumn] = [
            TextColumn("[cyan][{task.id}] [b blue]{task.fields[sha]}"),
            "[progress.percentage]{task.percentage:>3.0f}%",
            BarColumn(),
            "·",
            DownloadColumn(binary_units=True),
            "·",
            TransferSpeedColumn(),
            "·",
            TextColumn("[blue]{task.fields[data_node]}"),
        ]
        if self.config.download.show_filename:
            file_columns.extend(
                [
                    "·",
                    TextColumn("{task.fields[filename]}"),
                ]
            )
        file_progress = self.ui.make_progress(
            *file_columns,
            transient=True,
        )
        file_task_shas = {}
        start_callbacks = {}
        for file in queue:
            task_id = file_progress.add_task(
                "",
                total=file.size,
                visible=False,
                start=False,
                sha=short_sha(file.sha),
                filename=file.filename,
                data_node=file.data_node,
            )
            callback = partial(file_progress.start_task, task_id)
            file_task_shas[file.sha] = task_id
            start_callbacks[file.sha] = [callback]
        processor = Processor(
            config=self.config,
            fs=self.fs,
            files=queue,
            start_callbacks=start_callbacks,
        )
        if use_db:
            self.db.add(*processor.files)
        queue_size = len(processor.tasks)
        main_task_id = main_progress.add_task(
            "", total=queue_size, nb_errors=0
        )
        # TODO: rename ? installed/downloaded/completed/...
        files: list[File] = []
        errors: list[Err] = []
        remaining_dict = {file.sha: file for file in processor.files}
        try:
            with self.ui.live(
                file_progress,
                main_progress,
                disable=not show_progress,
            ) as live:
                async for result in self.iter_results(
                    processor,
                    file_progress,
                    file_task_shas,
                    live,
                ):
                    match result:
                        case Ok():
                            main_progress.update(main_task_id, advance=1)
                            file = result.data.file
                            file.status = FileStatus.Done
                            files.append(file)
                            emit(
                                Event.file_complete,
                                file=file,
                                destination=self.fs[file].drs,
                                start_time=result.data.start_time,
                                end_time=datetime.now(),
                            )
                            if file.dataset is not None:
                                is_dataset_complete = self.db.scalars(
                                    sql.dataset.is_complete(file.dataset)
                                )[0]
                                if is_dataset_complete:
                                    emit(
                                        Event.dataset_complete,
                                        dataset=file.dataset,
                                    )
                        case Err(_, err):
                            queue_size -= 1
                            result.data.file.status = FileStatus.Error
                            errors.append(result)
                            main_progress.update(
                                main_task_id,
                                total=queue_size,
                                nb_errors=len(errors),
                            )
                            emit(
                                Event.file_error,
                                file=result.data.file,
                                exception=err,
                            )

                    if use_db:
                        self.db.add(result.data.file)
                    remaining_dict.pop(result.data.file.sha, None)
        finally:
            if remaining_dict:
                logger.warning(f"Cancelling {len(remaining_dict)} downloads.")
                cancelled: list[File] = []
                for file in remaining_dict.values():
                    file.status = FileStatus.Cancelled
                    cancelled.append(file)
                    errors.append(Err(file, DownloadCancelled()))
                if use_db:
                    self.db.add(*cancelled)
        return files, errors

    def replace_queries(
        self,
        graph: Graph,
        mapping: tuple[str | None, str],
    ) -> None:
        to_replace = [
            q for q in graph.queries.values() if q.require == mapping[0]
        ]
        for query in to_replace:
            new_query = query.clone(compute_sha=False)
            new_query.require = mapping[1]
            new_query.compute_sha()
            graph.replace(query, new_query)
            self.replace_queries(graph, (query.sha, new_query.sha))

    def insert_default_query(self, *queries: Query) -> list[Query]:
        if self.config.api.default_query_id == "":
            return list(queries)
        default_query_id = self.config.api.default_query_id
        try:
            default_query = self.graph.get(default_query_id)
        except KeyError:
            raise UnknownDefaultQueryID(default_query_id)
        graph = Graph(None)
        graph.add(*queries)
        self.replace_queries(graph, (None, default_query.sha))
        return list(graph.queries.values())



from __future__ import annotations

from datetime import datetime

import click
from click.exceptions import Abort, Exit

from esgpull.cli.decorators import args, groups, opts
from esgpull.cli.utils import filter_keys, init_esgpull, parse_query, totable
from esgpull.context import IndexNode
from esgpull.exceptions import PageIndexError
from esgpull.graph import Graph
from esgpull.models import Query
from esgpull.tui import Verbosity


@click.command()
@args.facets
@groups.query_def
@groups.query_date
@groups.display
@groups.json_yaml
@opts.detail
@opts.no_default_query
@opts.show
@opts.dry_run
@opts.file
@opts.facets_hints
@opts.hints
@opts.yes
@opts.record
@opts.verbosity
def search(
    facets: list[str],
    ## query_def
    tags: list[str],
    require: str | None,
    distrib: str | None,
    latest: str | None,
    replica: str | None,
    retracted: str | None,
    ## query_date
    date_from: datetime | None,
    date_to: datetime | None,
    ## display
    _all: bool,
    zero: bool,
    page: int,
    ## json_yaml
    json: bool,
    yaml: bool,
    ## ungrouped
    detail: int | None,
    no_default_query: bool,
    show: bool,
    dry_run: bool,
    file: bool,
    facets_hints: bool,
    hints: list[str] | None,
    yes: bool,
    record: bool,
    verbosity: Verbosity,
) -> None:
    """
    Search datasets and files on ESGF

    More info
    """
    esg = init_esgpull(
        verbosity,
        safe=False,
        record=record,
        no_default_query=no_default_query,
    )
    with esg.ui.logging("search", onraise=Abort):
        query = parse_query(
            facets=facets,
            tags=tags,
            require=require,
            distrib=distrib,
            latest=latest,
            replica=replica,
            retracted=retracted,
        )
        query.compute_sha()
        esg.graph.resolve_require(query)
        query = esg.insert_default_query(query)[0]
        if show:
            if json:
                esg.ui.print(query.asdict(), json=True)
            elif yaml:
                esg.ui.print(query.asdict(), yaml=True)
            else:
                try:
                    graph = esg.graph.subgraph(query, parents=True)
                    esg.ui.print(graph)
                except KeyError:
                    esg.ui.print(query)
            esg.ui.raise_maybe_record(Exit(0))
        esg.graph.add(query, force=True)
        query = esg.graph.expand(query.sha)
        esg.context.probe()
        hits = esg.context.hits(
            query,
            file=file,
            date_from=date_from,
            date_to=date_to,
        )
        nb = sum(hits)
        page_size = esg.config.cli.page_size
        if detail is not None:
            page_size = 1
            page = detail
        nb_pages = (nb // page_size) or 1
        offset = page * page_size
        max_hits = min(page_size, nb - offset)
        if page > nb_pages:
            raise PageIndexError(page, nb_pages)
        elif zero:
            max_hits = 0
        elif _all:
            offset = 0
            max_hits = nb
        ids = range(offset, offset + max_hits)
        if dry_run:
            search_results = esg.context.prepare_search(
                query,
                file=file,
                hits=hits,
                offset=offset,
                max_hits=max_hits,
                date_from=date_from,
                date_to=date_to,
            )
            for result in search_results:
                esg.ui.print(result.request.url)
            esg.ui.raise_maybe_record(Exit(0))
        if facets_hints:
            not_distrib_query = query << Query(options=dict(distrib=False))
            index = IndexNode(esg.config.api.index_node)
            if index.is_bridge():
                first_file_result = esg.context.search_as_queries(
                    not_distrib_query,
                    file=True,
                    max_hits=1,
                    date_from=date_from,
                    date_to=date_to,
                )
                first_file = first_file_result[0].selection.asdict()
                esg.ui.print(list(first_file), json=True)
            else:
                facet_counts = esg.context.hints(
                    not_distrib_query,
                    file=file,
                    facets=["*"],
                    date_from=date_from,
                    date_to=date_to,
                )
                esg.ui.print(list(facet_counts[0]), json=True)
            esg.ui.raise_maybe_record(Exit(0))
        if hints is not None:
            facet_counts = esg.context.hints(
                query,
                file=file,
                facets=hints,
                date_from=date_from,
                date_to=date_to,
            )
            esg.ui.print(facet_counts, json=True)
            esg.ui.raise_maybe_record(Exit(0))
        if max_hits > 200 and not yes:
            nb_req = max_hits // esg.config.api.page_limit
            message = f"{nb_req} requests will be sent to ESGF. Send anyway?"
            if not esg.ui.ask(message, default=True):
                esg.ui.raise_maybe_record(Abort)
        if detail is not None:
            queries = esg.context.search_as_queries(
                query,
                file=file,
                hits=hits,
                offset=offset,
                max_hits=max_hits,
                keep_duplicates=True,
                date_from=date_from,
                date_to=date_to,
            )
            if json:
                selections = [q.selection.asdict() for q in queries]
                esg.ui.print(selections, json=True)
            elif yaml:
                selections = [q.selection.asdict() for q in queries]
                esg.ui.print(selections, yaml=True)
            else:
                graph = Graph(None)
                graph.add(*queries, clone=False)
                esg.ui.print(graph)
            esg.ui.raise_maybe_record(Exit(0))
        results = esg.context.search(
            query,
            file=file,
            hits=hits,
            offset=offset,
            max_hits=max_hits,
            keep_duplicates=True,
            date_from=date_from,
            date_to=date_to,
        )
        if json:
            esg.ui.print([f.asdict() for f in results], json=True)
        elif yaml:
            esg.ui.print([f.asdict() for f in results], yaml=True)
        else:
            f_or_d = "file" if file else "dataset"
            s = "s" if nb != 1 else ""
            esg.ui.print(f"Found {nb} {f_or_d}{s}.")
            if results:
                unique_ids = {r.master_id for r in results}
                unique_nodes = {(r.master_id, r.data_node) for r in results}
                needs_data_node = len(unique_nodes) > len(unique_ids)
                docs = filter_keys(results, ids=ids, data_node=needs_data_node)
                esg.ui.print(totable(docs))
        esg.ui.raise_maybe_record(Exit(0))



"""file_add_composite_index_dataset_id_status

Revision ID: d14f179e553c
Revises: e7edab5d4e4b
Create Date: 2025-06-18 16:05:35.721085

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'd14f179e553c'
down_revision = 'e7edab5d4e4b'
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('file', schema=None) as batch_op:
        batch_op.create_index('ix_file_dataset_status', ['dataset_id', 'status'], unique=False)

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('file', schema=None) as batch_op:
        batch_op.drop_index('ix_file_dataset_status')

    # ### end Alembic commands ###



from collections import Counter
from collections.abc import MutableMapping
from pathlib import Path

import click
import pyparsing as pp
import yaml
from click.exceptions import Abort, Exit
from rich.box import MINIMAL_DOUBLE_HEAD
from rich.prompt import Confirm
from rich.table import Table
from rich.text import Text

from esgpull.cli.decorators import opts
from esgpull.cli.utils import init_esgpull
from esgpull.graph import Graph
from esgpull.models import Options, Query, Tag
from esgpull.models.selection import FacetValues, Selection
from esgpull.tui import Verbosity, logger

SKIP = {"priority", "protocol"}
options_names = Options()._names
pp.ParserElement.set_default_whitespace_chars(" \t,")  # remove newline


def line(expr: pp.ParserElement) -> pp.ParserElement:
    return pp.LineStart().suppress() + expr + pp.LineEnd().suppress()


def sqbr(expr: pp.ParserElement) -> pp.ParserElement:
    return pp.Suppress("[") + expr + pp.Suppress("]")


anything = pp.Word(pp.printables)
word = pp.Word(pp.alphanums + "-_")
eq = pp.Suppress("=")

comment_start = pp.Char("#")[1, ...].suppress()
name_comment = pp.Group(line(comment_start + anything[1])[0, 1])
comment = line(comment_start + anything[...]).suppress()
facet_name = word("name")
facet_value = word
facet_values = pp.Group(word[1, ...])
facet = pp.Group(line(facet_name + eq + facet_values("vals")))("facet")
variable_cmip5 = pp.Group(
    line(
        "variable"
        + sqbr(facet_value)("realm")
        + sqbr(facet_value)("time_frequency")
        + eq
        + facet_values("variable")
    )
)("variable_cmip5")
variable_cmip6 = pp.Group(
    line(
        "variable"
        + sqbr(pp.Opt("table_id=") + facet_value("table_id"))
        + eq
        + facet_values("variable_id")
    )
)("variable_cmip6")
variable_no_sqbr = pp.Group(line("variable" + eq + facet_values("variable")))(
    "variable_no_sqbr"
)
otherwise = line(anything[...])("otherwise")
rest = pp.Group(
    comment
    | facet
    | variable_no_sqbr
    | variable_cmip5
    | variable_cmip6
    | otherwise
)[...]
selection_file = name_comment("name") + rest("rest")


def remove_duplicates(
    selection: MutableMapping[str, FacetValues],
) -> MutableMapping[str, FacetValues]:
    result: dict[str, FacetValues] = {}
    duplicates: dict[str, list[str]] = {}
    for name, values in selection.items():
        if isinstance(values, str):
            result[name] = values
            continue
        counter = Counter(values)
        nb_dup = sum(c > 1 for c in counter.values())
        if nb_dup == 0:
            result[name] = values
            continue
        duplicates[name] = list(dict(counter.most_common(nb_dup)))
    if duplicates:
        logger.warning(f"Duplicate values {duplicates}'")
    return result


def is_CMIP6(q: Query) -> bool:
    project = set(q.selection["project"] + q.selection["mip_era"])
    return "CMIP6" in project


def isnot_CMIP6(q: Query) -> bool:
    project = set(q.selection["project"] + q.selection["mip_era"])
    return any({"CMIP5", "CORDEX"} & project)


def fix_CMIP5(q: Query) -> Query:
    qd = q.asdict()
    if q.selection["frequency"]:
        qd["selection"]["time_frequency"] = qd["selection"].pop("frequency")
    return Query(**qd)


def convert_file(path: Path) -> Graph:
    logger.info(path)
    query = Query()
    kids: list[Query] = []
    result = selection_file.parse_file(path)
    if result.name:
        # name = result.name[0].split("@", 1)[0]
        query.tags.append(Tag(name=result.name[0]))
    for line in result.rest:
        if line.facet:
            name, values = line.facet.as_list()
            if name in SKIP:
                continue
            elif name in options_names:
                if len(values) > 1:
                    raise ValueError({name: values})
                d = {name: values[0]}
                logger.debug(f"OPTION {d}")
                setattr(query.options, name, values[0])
            elif name in Selection._facet_names:
                d = {name: values}
                logger.debug(f"FACET {d}")
                query.selection[name] = list(set(values))
            else:
                raise ValueError(f"{name!r} undefined\n{path.read_text()}")
        elif line.variable_cmip5:
            selection = line.variable_cmip5.as_dict()
            logger.debug(f"SUBQUERY {selection}")
            kid = Query(selection=remove_duplicates(selection), tracked=True)
            kids.append(kid)
        elif line.variable_cmip6:
            selection = line.variable_cmip6.as_dict()
            project = query.selection["project"]
            if project and project[0] in ["CMIP5", "CORDEX"]:
                if "variable_id" in selection:
                    selection["variable"] = selection.pop("variable_id")
                if "table_id" in selection:
                    selection["time_frequency"] = selection.pop("table_id")
            logger.debug(f"SUBQUERY {selection}")
            kid = Query(selection=remove_duplicates(selection))
            kids.append(kid)
        elif line.variable_no_sqbr:
            logger.error(line.as_dict())
        elif line.as_list():
            logger.error(line.as_list())
    if len(kids) > 1:
        query.untrack()
    elif len(kids) == 1:
        query = query << kids.pop()
    else:
        query.track(query.options)
    if isnot_CMIP6(query):
        query = fix_CMIP5(query)
    query.compute_sha()
    graph = Graph(None)
    graph.add(query)
    for kid in kids:
        if isnot_CMIP6(query) or isnot_CMIP6(kid):
            kid = fix_CMIP5(kid)
        kid.require = query.sha
        kid.compute_sha()
        expanded = query << kid
        kid.track(expanded.options)
        graph.add(kid)
    return graph


CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument(
    "paths",
    required=True,
    type=click.Path(exists=True, path_type=Path),
    nargs=-1,
)
@click.option(
    "--out",
    "-o",
    type=click.Path(exists=False, path_type=Path),
    default=None,
)
@click.option("print_table", "--table", is_flag=True, default=False)
@click.option("print_graph", "--graph", is_flag=True, default=False)
@opts.record
@opts.verbosity
def convert(
    paths: list[Path],
    out: Path | None,
    print_table: bool,
    print_graph: bool,
    record: bool,
    verbosity: Verbosity,
):
    """
    Convert synda selection files to esgpull queries

    The flags `--table/--graph` are used to select the console output design.

    Use `--out <path/to/output.yaml>` to generate a query file containing the converted synda
    selection files. This file can be used as an input to the `add --query-file` command.

    Note that `convert` takes any number of input paths, and produces a single output.
    Using `**/*` as the input path is a good way to convert a whole tree of selection files.

    As a an arbitrary convention, the generated queries will be tagged with the content of the first line of a synda selection file, if that line starts with `#` and has a single word (+symbols) without whitespaces in it.
    """
    esg = init_esgpull(verbosity, safe=False, record=record)
    with esg.ui.logging("convert", onraise=Abort):
        if len(paths) == 0:
            esg.ui.print("No file provided")
            raise click.exceptions.Exit(0)
        if out is not None:
            out_str = f"[yellow]{out.resolve()}[/]"
            if out.is_file():
                esg.ui.print(f"Overwriting existing file {out_str}")
                allow = Confirm.ask("Continue?", default=False)
                if not allow:
                    raise click.exceptions.Exit(0)
        if out is None and not print_table and not print_graph:
            esg.ui.print(
                "[red]Error[/]: At least one of "
                "[yellow]--out/--table/--graph[/] is required."
            )
            raise click.exceptions.Exit(1)
        table = Table(
            box=MINIMAL_DOUBLE_HEAD,
            show_edge=False,
            show_lines=True,
        )
        table.add_column(Text("path", justify="left"))
        table.add_column(Text("file", justify="center"))
        table.add_column(Text("query", justify="center"))
        full_graph = Graph(None)
        empty_query = Query()
        empty_query.options.apply_defaults(empty_query.options)
        empty_query.compute_sha()
        for path in paths:
            if path.is_file():
                try:
                    graph = convert_file(path)
                except Exception as e:
                    msg = f"Cannot convert {path.resolve()}"
                    logger.exception(e)
                    logger.error(msg)
                    esg.ui.print(f"[red]Error[/]: {msg}")
                    continue
                queries = list(graph.queries.values())
                if len(queries) == 1 and queries[0].sha == empty_query.sha:
                    esg.ui.print(
                        f"Skipping empty query produced from {path.resolve()}"
                    )
                    continue
                path_text = Text(str(path).replace("/", "/\n"), style="yellow")
                file_text = Text(path.read_text())
                table.add_row(path_text, file_text, graph)
                full_graph.add(*queries, force=True)
        if print_table:
            esg.ui.print(table)
        if print_graph:
            esg.ui.print(full_graph)
        if out is not None:
            graph_as_yaml = yaml.dump(full_graph.dump())
            try:
                with out.open("w") as f:
                    f.write(graph_as_yaml)
                esg.ui.print(f":+1: [green]Graph was written to[/] {out_str}")
            except Exception:
                out.unlink()
                raise
        esg.ui.raise_maybe_record(Exit(0))



# from esgpull.constants import ROOT_ENV


class EsgpullException(Exception):
    msg: str = NotImplemented

    def __init__(self, *args, **kwargs):
        self.message = self.msg.format(*args, **kwargs)
        super().__init__(self.message.strip())


# class NoRootError(EsgpullException):
#     msg = f"Environment variable `{ROOT_ENV}` must be set."


# class UnknownFacet(EsgpullException):
#     msg = "{}"


class FacetNameError(EsgpullException, AttributeError):
    """
    AttributeError is required for autocomplete engines (e.g. jupyter).
    """

    msg = "'{}' is not a valid facet."


class AlreadySetFacet(EsgpullException):
    msg = "'{}' is already set to [{}]"


class DuplicateFacet(EsgpullException):
    msg = "'{}:{}'\n{}"


class QueryDuplicate(EsgpullException):
    msg = "{}"


class PageIndexError(EsgpullException, IndexError):
    msg = "Cannot show page {}/{}."


# # errors meant for use when validation is implemented
# class UnknownFacetValue(EsgpullException):
#     msg = "'{}' is not valid for {}."
# class ImpossibleFacet(EsgpullException):
#     msg = """Facet '{}' is not available with this query:
#     {}"""


class SolrUnstableQueryError(EsgpullException):
    msg = """
    Solr can not handle this query:
    {}
    """


class QuerySourceError(EsgpullException):
    msg = """
    This source cannot be parsed as a query:
    {}
    """


class TooShortKeyError(EsgpullException, KeyError):
    msg = "{}"


class GraphWithoutDatabase(EsgpullException):
    msg = "Graph is not connected to a database."


class DownloadKindError(EsgpullException):
    msg = """{} is not a valid download kind. Choose from:
    * Download
    * ChunkedDownload
    * MultiSourceChunkedDownload
    """


class DownloadSizeError(EsgpullException):
    msg = """
    Downloaded file is larger than expected: {} > {}
    """


class DownloadCancelled(EsgpullException):
    msg = """
    Download cancelled by user.
    """


class NoClauseError(EsgpullException):
    msg = """
    No clause provided (query might be empty).
    """


class BadConfigError(EsgpullException):
    msg = """
    Please fix your config, located at {}
    """


class VirtualConfigError(EsgpullException):
    msg = """
    This config was not loaded from a file.
    """


class InstallException(EsgpullException): ...


class UnknownDefaultQueryID(EsgpullException):
    msg = "{}"


class UntrackableQuery(EsgpullException):
    msg = """
    {} cannot be tracked, it has unset options.
    """


class UnsetOptionsError(EsgpullException):
    msg = """
    {} has some unset options.
    """


class UnregisteredInstallPath(InstallException):
    msg = "{}"


class UnknownInstallName(InstallException):
    msg = "{!r}"


class NoInstallPath(InstallException):
    msg = """Choose or install one

Show existing install locations with:
$ esgpull self choose

Install a new location with:
$ esgpull self install
"""


class InvalidInstallPath(InstallException):
    msg = """{path}

Choose this install location with:
$ esgpull self choose {path}

Install a new location with:
$ esgpull self install
"""


class AlreadyInstalledPath(InstallException):
    msg = """{path}
{msg}
"""


class AlreadyInstalledName(InstallException):
    msg = """{name}
{msg}
"""



# from math import ceil
from collections.abc import AsyncGenerator
from dataclasses import dataclass
from datetime import datetime

from httpx import AsyncClient

from esgpull.fs import Digest
from esgpull.models import File

# import asyncio
# from urllib.parse import urlsplit
# from esgpull.auth import Auth
# from esgpull.context import Context


@dataclass
class DownloadCtx:
    file: File
    completed: int = 0
    chunk: bytes | None = None
    digest: Digest | None = None
    start_time: datetime | None = None

    @property
    def finished(self) -> bool:
        return self.completed == self.file.size

    @property
    def error(self) -> bool:
        return self.completed > self.file.size

    def update_digest(self) -> None:
        if self.digest is not None and self.chunk is not None:
            self.digest.update(self.chunk)


class BaseDownloader:
    def stream(
        self,
        client: AsyncClient,
        ctx: DownloadCtx,
        chunk_size: int,
    ) -> AsyncGenerator[DownloadCtx, None]:
        raise NotImplementedError


class Simple(BaseDownloader):
    """
    Simple chunked async downloader.
    """

    async def stream(
        self,
        client: AsyncClient,
        ctx: DownloadCtx,
        chunk_size: int,
    ) -> AsyncGenerator[DownloadCtx, None]:
        ctx.start_time = datetime.now()
        async with client.stream("GET", ctx.file.url) as resp:
            resp.raise_for_status()
            async for chunk in resp.aiter_bytes(chunk_size=chunk_size):
                ctx.completed += len(chunk)
                ctx.chunk = chunk
                ctx.update_digest()
                yield ctx


# class Distributed(BaseDownloader):
#     """
#     Distributed chunked async downloader.
#     Fetches chunks from multiple URLs pointing to the same file.
#     """

#     def __init__(
#         self,
#         auth: Auth,
#         *,
#         file: File | None = None,
#         url: str | None = None,
#         config: Config | None = None,
#         max_ping: float = 5.0,
#     ) -> None:
#         super().__init__(auth, file=file, url=url, config=config)
#         self.max_ping = max_ping

#     async def try_url(self, url: str, client: AsyncClient) -> str | None:
#         result = None
#         node = urlsplit(url).netloc
#         print(f"trying url on '{node}'")
#         try:
#             resp = await client.head(url)
#             print(f"got response on '{node}'")
#             resp.raise_for_status()
#             accept_ranges = resp.headers.get("Accept-Ranges")
#             content_length = resp.headers.get("Content-Length")
#             if (
#                 accept_ranges == "bytes"
#                 and int(content_length) == self.file.size
#             ):
#                 result = str(resp.url)
#             else:
#                 print(dict(resp.headers))
#         except HTTPError as err:
#             print(type(err))
#             print(err.request.headers)
#         return result

#     async def process_queue(
#         self, url: str, queue: asyncio.Queue
#     ) -> tuple[list[tuple[int, bytes]], str]:
#         node = urlsplit(url).netloc
#         print(f"starting process on '{node}'")
#         chunks: list[tuple[int, bytes]] = []
#         async with self.make_client() as client:
#             final_url = await self.try_url(url, client)
#             if final_url is None:
#                 print(f"no url found for '{node}'")
#                 return chunks, url
#             else:
#                 url = final_url
#             while not queue.empty():
#                 chunk_idx = await queue.get()
#                 print(f"processing chunk {chunk_idx} on '{node}'")
#                 start = chunk_idx * self.config.download.chunk_size
#                 end = min(
#                     self.file.size,
#                     (chunk_idx + 1) * self.config.download.chunk_size - 1,
#                 )
#                 headers = {"Range": f"bytes={start}-{end}"}
#                 resp = await client.get(url, headers=headers)
#                 queue.task_done()
#                 if resp.status_code == 206:
#                     chunks.append((chunk_idx, resp.content))
#                 else:
#                     await queue.put(chunk_idx)
#                     print(f"error status {resp.status_code} on '{node}'")
#                     break
#         return chunks, url

#     async def fetch_urls(self) -> list[str]:
#         ctx = Context(distrib=True)
#         ctx.query.instance_id = self.file.file_id
#         results = await ctx._search(file=True)
#         files = [File.from_dict(item) for item in results]
#         return [file.url for file in files]

# async def aget(self) -> bytes:
#     nb_chunks = ceil(self.file.size / self.config.download.chunk_size)
#     queue: asyncio.Queue[int] = asyncio.Queue(nb_chunks)
#     for chunk_idx in range(nb_chunks):
#         queue.put_nowait(chunk_idx)
#     completed: list[bool] = [False for _ in range(nb_chunks)]
#     chunks: list[bytes] = [bytes() for _ in range(nb_chunks)]
#     urls = await self.fetch_urls()
#     workers = [self.process_queue(url, queue) for url in urls]
#     for future in asyncio.as_completed(workers):
#         some_chunks, url = await future
#         print(f"got {len(some_chunks)} chunks from {url}")
#         for chunk_idx, chunk in some_chunks:
#             completed[chunk_idx] = True
#             chunks[chunk_idx] = chunk
#     if not all(completed):
#         raise ValueError("TODO: progressive write (with .part file)")
#     return b"".join(chunks)



import pytest

from esgpull.database import Database
from esgpull.models import Dataset, DatasetRecord, File


@pytest.fixture
def db(config):
    return Database.from_config(config)


def test_dataset_record_serialize():
    """Test DatasetRecord can parse ESGF API response."""
    source = {
        "instance_id": "CMIP6.CMIP.NCAR.CESM2.historical.r1i1p1f1.Amon.tas.gn.v20190308|esgf-data.ucar.edu",
        "data_node": "esgf-data.ucar.edu",
        "size": 123456789,
        "number_of_files": 10,
    }

    record = DatasetRecord.serialize(source)

    assert (
        record.dataset_id
        == "CMIP6.CMIP.NCAR.CESM2.historical.r1i1p1f1.Amon.tas.gn.v20190308"
    )
    assert (
        record.master_id
        == "CMIP6.CMIP.NCAR.CESM2.historical.r1i1p1f1.Amon.tas.gn"
    )
    assert record.version == "v20190308"
    assert record.data_node == "esgf-data.ucar.edu"
    assert record.size == 123456789
    assert record.number_of_files == 10


def test_dataset_file_relationship(db):
    """Test the relationship between Dataset and File models."""
    # Create a dataset
    dataset_id = (
        "CMIP6.CMIP.NCAR.CESM2.historical.r1i1p1f1.Amon.tas.gn.v20190308"
    )
    dataset = Dataset(
        dataset_id=dataset_id,
        total_files=2,
    )
    db.session.add(dataset)

    # Create files
    file1 = File(
        file_id=f"{dataset_id}.nc_1",
        dataset_id=dataset_id,
        master_id="master",
        url="http://example.com/file1.nc",
        version="v20190308",
        filename="file1.nc",
        local_path="/data/file1.nc",
        data_node="example.com",
        checksum="checksum1",
        checksum_type="sha256",
        size=1000,
    )
    file1.compute_sha()
    db.session.add(file1)

    db.session.commit()

    # Test relationship
    assert len(dataset.files) == 1
    assert dataset.files[0].file_id == file1.file_id
    assert file1.dataset == dataset



import os

CONFIG_FILENAME = "config.toml"
INSTALLS_PATH_ENV = "ESGPULL_INSTALLS_PATH"
ROOT_ENV = "ESGPULL_CURRENT"
ESGPULL_DEBUG = os.environ.get("ESGPULL_DEBUG", "0") == "1"
ESGPULL_DEBUG_LOCALS = os.environ.get("ESGPULL_DEBUG_LOCALS", "0") == "1"

IDP = "/esgf-idp/openid/"
CEDA_IDP = "/OpenID/Provider/server/"
PROVIDERS = {
    "esg-dn1.nsc.liu.se": IDP,
    "esgf-data.dkrz.de": IDP,
    "ceda.ac.uk": CEDA_IDP,
    "esgf-node.ipsl.upmc.fr": IDP,
    "esgf-node.llnl.gov": IDP,
    "esgf.nci.org.au": IDP,
}


DEFAULT_FACETS = [
    "project",
    "mip_era",
    "experiment",
    "experiment_id",
    "institute",
    "institution_id",
    "model",
    "table_id",
    "activity_id",
    "ensemble",
    "variant_label",
    "realm",
    "frequency",
    "time_frequency",
    "variable",
    "variable_id",
    "dataset_id",
    "source_id",
    "domain",
    "driving_model",
    "rcm_name",
    "member_id",
    "cmor_table",
]
EXTRA_FACETS = [
    "query",
    "start",
    "end",
    # "fields",
    "facets",
    "url",
    "data_node",
    "index_node",
    "master_id",
    "instance_id",  # search does not work with instance_id
    "title",
    "variable_long_name",
    "experiment_family",
]
DEFAULT_CONSTRAINTS_WITH_VALUE: dict[str, str] = {}



from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime, timezone

import click
from click.exceptions import Abort, Exit

from esgpull.cli.decorators import args, opts
from esgpull.cli.utils import get_queries, init_esgpull, valid_name_tag
from esgpull.context import HintsDict, ResultSearch
from esgpull.exceptions import UnsetOptionsError
from esgpull.models import Dataset, File, FileStatus, Query
from esgpull.tui import Verbosity
from esgpull.utils import format_size


@dataclass
class QueryFiles:
    query: Query
    expanded: Query
    skip: bool = False
    datasets: list[Dataset] = field(default_factory=list)
    files: list[File] = field(default_factory=list)
    dataset_hits: int = field(init=False)
    hits: int = field(init=False)
    hints: HintsDict = field(init=False)
    results: list[ResultSearch] = field(init=False)
    dataset_results: list[ResultSearch] = field(init=False)


@click.command()
@args.query_id
@opts.tag
@opts.children
@opts.yes
@opts.record
@opts.verbosity
def update(
    query_id: str | None,
    tag: str | None,
    children: bool,
    yes: bool,
    record: bool,
    verbosity: Verbosity,
) -> None:
    """
    Fetch files, link files <-> queries, send files to download queue
    """
    esg = init_esgpull(verbosity, record=record)
    with esg.ui.logging("update", onraise=Abort):
        # Select which queries to update + setup
        if query_id is None and tag is None:
            esg.graph.load_db()
            queries = list(esg.graph.queries.values())
        else:
            if not valid_name_tag(esg.graph, esg.ui, query_id, tag):
                esg.ui.raise_maybe_record(Exit(1))
            queries = get_queries(
                esg.graph,
                query_id,
                tag,
                children=children,
            )
        qfs: list[QueryFiles] = []
        for query in queries:
            expanded = esg.graph.expand(query.sha)
            if query.tracked and not expanded.trackable():
                esg.ui.print(query)
                raise UnsetOptionsError(query.name)
            elif query.tracked:
                qfs.append(QueryFiles(query, expanded))
        queries = [
            query
            for query in queries
            if query.tracked and query.sha != "LEGACY"
        ]
        if not qfs:
            esg.ui.print(":stop_sign: Trying to update untracked queries.")
            esg.ui.raise_maybe_record(Exit(0))
        esg.context.probe()
        hints = esg.context.hints(
            *[qf.expanded for qf in qfs],
            file=True,
            facets=["index_node"],
        )
        dataset_hits = esg.context.hits(
            *[qf.expanded for qf in qfs],
            file=False,
        )
        for qf, qf_hints, qf_dataset_hits in zip(qfs, hints, dataset_hits):
            qf.hits = sum(esg.context.hits_from_hints(qf_hints))
            if qf_hints:
                qf.hints = qf_hints
                qf.dataset_hits = qf_dataset_hits
            else:
                qf.skip = True
        for qf in qfs:
            s = "s" if qf.hits > 1 else ""
            esg.ui.print(
                f"{qf.query.rich_name} -> {qf.hits} file{s} (before replica de-duplication)."
            )
        total_hits = sum([qf.hits for qf in qfs])
        if total_hits == 0:
            esg.ui.print("No files found.")
            esg.ui.raise_maybe_record(Exit(0))
        elif len(qfs) > 1:
            esg.ui.print(f"{total_hits} files found.")
        qfs = [qf for qf in qfs if not qf.skip]
        # Prepare optimally distributed requests to ESGF
        # [?] TODO: fetch FastFile first to determine what to fetch in detail later
        #   It might be interesting for the special case where all files already
        #   exist in db, then the detailed fetch could be skipped.
        for qf in qfs:
            qf_dataset_results = esg.context.prepare_search(
                qf.expanded,
                file=False,
                hits=[qf.dataset_hits],
                max_hits=None,
            )
            if esg.config.api.use_custom_distribution_algorithm:
                qf_results = esg.context.prepare_search_distributed(
                    qf.expanded,
                    file=True,
                    hints=[qf.hints],
                    max_hits=None,
                )
            else:
                qf_results = esg.context.prepare_search(
                    qf.expanded,
                    file=True,
                    hits=[qf.hits],
                    max_hits=None,
                )
            nb_req = len(qf_dataset_results) + len(qf_results)
            if nb_req > 50:
                msg = (
                    f"{nb_req} requests will be sent to ESGF to"
                    f" update {qf.query.rich_name}. Send anyway?"
                )
                match esg.ui.choice(msg, ["y", "n", "u"], default="n"):
                    case "u":
                        esg.ui.print(f"{qf.query.rich_name} is now untracked.")
                        qf.query.tracked = False
                        qf_results = []
                        qf_dataset_results = []
                    case "n":
                        qf_results = []
                        qf_dataset_results = []
                    case _:
                        ...
            qf.results = qf_results
            qf.dataset_results = qf_dataset_results
        # Fetch files and update db
        # [?] TODO: dry_run to print urls here
        with esg.ui.spinner("Fetching datasets"):
            coros = []
            for qf in qfs:
                coro = esg.context._datasets(
                    *qf.dataset_results, keep_duplicates=False
                )
                coros.append(coro)
            datasets = esg.context.sync_gather(*coros)
            for qf, qf_datasets in zip(qfs, datasets):
                qf.datasets = [
                    Dataset(
                        dataset_id=record.dataset_id,
                        total_files=record.number_of_files,
                    )
                    for record in qf_datasets
                ]
        with esg.ui.spinner("Fetching files"):
            coros = []
            for qf in qfs:
                coro = esg.context._files(*qf.results, keep_duplicates=False)
                coros.append(coro)
            files = esg.context.sync_gather(*coros)
            for qf, qf_files in zip(qfs, files):
                qf.files = qf_files
        for qf in qfs:
            if not qf.query.tracked:
                esg.db.add(qf.query)
                continue
            with esg.db.commit_context():
                unregistered_datasets = [
                    f for f in qf.datasets if f not in esg.db
                ]
                if len(unregistered_datasets) > 0:
                    esg.ui.print(
                        f"Adding {len(unregistered_datasets)} new datasets to database."
                    )
                    esg.db.session.add_all(unregistered_datasets)
                files_from_db = [
                    esg.db.get(File, f.sha) for f in qf.files if f in esg.db
                ]
                registered_files = [f for f in files_from_db if f is not None]
                unregistered_files = [f for f in qf.files if f not in esg.db]
                if len(unregistered_files) > 0:
                    esg.ui.print(
                        f"Adding {len(unregistered_files)} new files to database."
                    )
                    esg.db.session.add_all(unregistered_files)
                files = registered_files + unregistered_files
            not_done_files = [
                file for file in files if file.status != FileStatus.Done
            ]
            download_size = sum(file.size for file in not_done_files)
            msg = (
                f"\n{qf.query.rich_name}: {len(not_done_files)} "
                f" files ({format_size(download_size)}) to download."
                f"\nLink to query and send to download queue?"
            )
            if yes:
                choice = "y"
            else:
                while (
                    choice := esg.ui.choice(
                        msg,
                        choices=["y", "n", "show"],
                        show_choices=True,
                    )
                ) and choice == "show":
                    esg.ui.print(esg.graph.subgraph(qf.query, parents=True))
            if choice == "y":
                legacy = esg.legacy_query
                has_legacy = legacy.state.persistent
                applied_changes = False
                with esg.db.commit_context():
                    for file in esg.ui.track(
                        files,
                        description=f"{qf.query.rich_name}",
                    ):
                        if file.status != FileStatus.Done:
                            file.status = FileStatus.Queued
                        if has_legacy and legacy in file.queries:
                            _ = esg.db.unlink(query=legacy, file=file)
                        changed = esg.db.link(query=qf.query, file=file)
                        applied_changes = applied_changes or changed
                    if applied_changes:
                        qf.query.updated_at = datetime.now(timezone.utc)
                    esg.db.session.add(qf.query)
        esg.ui.raise_maybe_record(Exit(0))



"""update tables

Revision ID: 0.3.0
Revises:
Create Date: 2023-02-08 14:35:52.386481

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "0.3.0"
down_revision = None
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "facet",
        sa.Column("name", sa.String(length=64), nullable=False),
        sa.Column("value", sa.String(length=255), nullable=False),
        sa.Column("sha", sa.String(length=40), nullable=False),
        sa.PrimaryKeyConstraint("sha"),
    )
    op.create_table(
        "file",
        sa.Column("file_id", sa.String(length=255), nullable=False),
        sa.Column("dataset_id", sa.String(length=255), nullable=False),
        sa.Column("master_id", sa.String(length=255), nullable=False),
        sa.Column("url", sa.String(length=255), nullable=False),
        sa.Column("version", sa.String(length=16), nullable=False),
        sa.Column("filename", sa.String(length=255), nullable=False),
        sa.Column("local_path", sa.String(length=255), nullable=False),
        sa.Column("data_node", sa.String(length=40), nullable=False),
        sa.Column("checksum", sa.String(length=64), nullable=False),
        sa.Column("checksum_type", sa.String(length=16), nullable=False),
        sa.Column("size", sa.Integer(), nullable=False),
        sa.Column(
            "status",
            sa.Enum(
                "New",
                "Queued",
                "Starting",
                "Started",
                "Pausing",
                "Paused",
                "Error",
                "Cancelled",
                "Done",
                name="filestatus",
            ),
            nullable=False,
        ),
        sa.Column("sha", sa.String(length=40), nullable=False),
        sa.PrimaryKeyConstraint("sha"),
        sa.UniqueConstraint("file_id"),
    )
    op.create_table(
        "options",
        sa.Column(
            "distrib",
            sa.Enum("false", "true", "none", "notset", name="option"),
            nullable=False,
        ),
        sa.Column(
            "latest",
            sa.Enum("false", "true", "none", "notset", name="option"),
            nullable=False,
        ),
        sa.Column(
            "replica",
            sa.Enum("false", "true", "none", "notset", name="option"),
            nullable=False,
        ),
        sa.Column(
            "retracted",
            sa.Enum("false", "true", "none", "notset", name="option"),
            nullable=False,
        ),
        sa.Column("sha", sa.String(length=40), nullable=False),
        sa.PrimaryKeyConstraint("sha"),
    )
    op.create_table(
        "selection",
        sa.Column("sha", sa.String(length=40), nullable=False),
        sa.PrimaryKeyConstraint("sha"),
    )
    op.create_table(
        "tag",
        sa.Column("name", sa.String(length=255), nullable=False),
        sa.Column("description", sa.Text(), nullable=True),
        sa.Column("sha", sa.String(length=40), nullable=False),
        sa.PrimaryKeyConstraint("sha"),
    )
    op.create_table(
        "query",
        sa.Column("tracked", sa.Boolean(), nullable=False),
        sa.Column("require", sa.String(length=40), nullable=True),
        sa.Column("options_sha", sa.String(length=40), nullable=False),
        sa.Column("selection_sha", sa.String(length=40), nullable=False),
        sa.Column("sha", sa.String(length=40), nullable=False),
        sa.ForeignKeyConstraint(
            ["options_sha"],
            ["options.sha"],
        ),
        sa.ForeignKeyConstraint(
            ["selection_sha"],
            ["selection.sha"],
        ),
        sa.PrimaryKeyConstraint("sha"),
    )
    op.create_table(
        "selection_facet",
        sa.Column("selection_sha", sa.String(length=40), nullable=False),
        sa.Column("facet_sha", sa.String(length=40), nullable=False),
        sa.ForeignKeyConstraint(
            ["facet_sha"],
            ["facet.sha"],
        ),
        sa.ForeignKeyConstraint(
            ["selection_sha"],
            ["selection.sha"],
        ),
        sa.PrimaryKeyConstraint("selection_sha", "facet_sha"),
    )
    op.create_table(
        "query_file",
        sa.Column("query_sha", sa.String(length=40), nullable=False),
        sa.Column("file_sha", sa.String(length=40), nullable=False),
        sa.ForeignKeyConstraint(
            ["file_sha"],
            ["file.sha"],
        ),
        sa.ForeignKeyConstraint(
            ["query_sha"],
            ["query.sha"],
        ),
        sa.PrimaryKeyConstraint("query_sha", "file_sha"),
    )
    op.create_table(
        "query_tag",
        sa.Column("query_sha", sa.String(length=40), nullable=False),
        sa.Column("tag_sha", sa.String(length=40), nullable=False),
        sa.ForeignKeyConstraint(
            ["query_sha"],
            ["query.sha"],
        ),
        sa.ForeignKeyConstraint(
            ["tag_sha"],
            ["tag.sha"],
        ),
        sa.PrimaryKeyConstraint("query_sha", "tag_sha"),
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("query_tag")
    op.drop_table("query_file")
    op.drop_table("selection_facet")
    op.drop_table("query")
    op.drop_table("tag")
    op.drop_table("selection")
    op.drop_table("options")
    op.drop_table("file")
    op.drop_table("facet")
    # ### end Alembic commands ###



from collections import Counter
from collections.abc import Sequence

import click
from click.exceptions import Abort, Exit

from esgpull.cli.decorators import args, opts
from esgpull.cli.utils import init_esgpull
from esgpull.models import FileStatus, sql
from esgpull.tui import Verbosity


@click.command()
@args.status
@opts.verbosity
def retry(
    status: Sequence[FileStatus],
    verbosity: Verbosity,
):
    """
    Re-queue failed and cancelled downloads
    """
    if not status:
        status = FileStatus.retryable()
    esg = init_esgpull(verbosity)
    with esg.ui.logging("retry", onraise=Abort):
        assert FileStatus.Done not in status
        assert FileStatus.Queued not in status
        files = list(esg.db.scalars(sql.file.with_status(*status)))
        status_str = "/".join(f"[bold red]{s.value}[/]" for s in status)
        if not files:
            esg.ui.print(f"No {status_str} files found.")
            raise Exit(0)
        counts = Counter(file.status for file in files)
        for file in files:
            file.status = FileStatus.Queued
        esg.db.add(*files)
        msg = "Sent back to the queue: "
        msg += ", ".join(
            f"{count} [bold red]{status.value}[/]"
            for status, count in counts.items()
        )
        esg.ui.print(msg)



from __future__ import annotations

from pathlib import Path

import click
from click.exceptions import Abort, Exit

from esgpull.cli.decorators import args, groups, opts
from esgpull.cli.utils import (
    init_esgpull,
    parse_query,
    serialize_queries_from_file,
)
from esgpull.graph import Graph
from esgpull.models import Query
from esgpull.tui import Verbosity


@click.command()
@args.facets
@groups.query_def
@opts.query_file
@opts.track
@opts.no_default_query
@opts.record
@opts.verbosity
def add(
    facets: list[str],
    ## query_def
    tags: list[str],
    require: str | None,
    distrib: str | None,
    latest: str | None,
    replica: str | None,
    retracted: str | None,
    ## ungrouped
    query_file: Path | None,
    track: bool,
    no_default_query: bool,
    record: bool,
    verbosity: Verbosity,
) -> None:
    """
    Add queries to the database

    OPTIONS / FACETS examples:

        esgpull add --distrib true variable_id:co2,co3 mip_era:CMIP6

            Syntax reference: http://www.esgf.io/esgf-download/search/

        esgpull add --query-file path/to/query.yaml

            Valid query files are usually created with either `show --json/--yaml` or `convert` commands.

    Queries are `untracked` by default.

    To fetch files from ESGF and link them to a query, see the `track` and `update` commands.
    """
    esg = init_esgpull(
        verbosity,
        record=record,
        no_default_query=no_default_query,
    )
    with esg.ui.logging("add", onraise=Abort):
        if query_file is not None:
            queries = serialize_queries_from_file(query_file)
        else:
            query = parse_query(
                facets=facets,
                tags=tags,
                require=require,
                distrib=distrib,
                latest=latest,
                replica=replica,
                retracted=retracted,
            )
            esg.graph.resolve_require(query)
            if track:
                if query.require is not None:
                    expanded = esg.graph.expand(query.require)
                else:
                    expanded = query
                query.track(expanded.options)
            queries = [query]
        queries = esg.insert_default_query(*queries)
        subgraph = Graph(None)
        subgraph.add(*queries)
        esg.ui.print(subgraph)
        empty = Query()
        empty.compute_sha()
        next_steps_queries = []
        for query in queries:
            query.compute_sha()
            esg.graph.resolve_require(query)
            if query.sha == empty.sha:
                esg.ui.print(":stop_sign: Trying to add empty query.")
                esg.ui.raise_maybe_record(Exit(1))
            if query.sha in esg.graph:  # esg.graph.has(sha=query.sha):
                esg.ui.print(f"Skipping existing query: {query.rich_name}")
            else:
                esg.graph.add(query)
                if query.tracked:
                    next_steps_queries.append(query)
        new_queries = esg.graph.merge()
        nb = len(new_queries)
        ies = "ies" if nb > 1 else "y"
        if new_queries:
            esg.ui.print(f":+1: {nb} new quer{ies} added.")
        else:
            esg.ui.print(":stop_sign: No new query was added.")
        if next_steps_queries:
            esg.ui.print("\nNext steps:\n")
            for query in next_steps_queries:
                esg.ui.print(f"\tesgpull update {query.sha[:6]}")
        esg.ui.raise_maybe_record(Exit(0))



# Sample plugin for testing
# This plugin defines handlers for all supported event types

from esgpull.plugin import Event, on

# Version compatibility
MIN_ESGPULL_VERSION = "0.8.0"
MAX_ESGPULL_VERSION = None


# Configuration class
class Config:
    """Configuration for sample_plugin"""

    log_level = "INFO"
    notification_method = "console"
    max_retries = 3


# Track calls for testing
calls = {"file_complete": 0, "file_error": 0, "dataset_complete": 0}


@on(Event.file_complete, priority="normal")
def handle_file_complete(file, destination, start_time, end_time, logger):
    """Handle file complete event"""
    logger.info(f"File downloaded: {file.filename}")
    calls["file_complete"] += 1


@on(Event.file_error, priority="high")
def handle_file_error(file, exception, logger):
    """Handle file error event"""
    logger.error(f"Download failed for {file.filename}: {exception}")
    calls["file_error"] += 1


@on(Event.dataset_complete, priority="low")
def handle_dataset_complete(dataset, logger):
    """Handle dataset_complete updated event"""
    logger.info(f"Dataset downloaded: {dataset.dataset_id}")
    calls["dataset_complete"] += 1



import pytest

from esgpull.cli.utils import parse_facets
from esgpull.models import Selection

test_cases = [
    (["ocean", "temperature"], Selection(query=["ocean", "temperature"])),
    (["ocean temperature"], Selection(query='"ocean temperature"')),
    (
        [
            "project:CMIP6",
            "experiment_id:ssp126,ssp245,ssp585",
            "variable_id:hurt,lai,pr,sfcWind,tasmax,tasmin,ts,tos",
            "frequency:day",
            "nominal_resolution:100 km",
        ],
        Selection(
            project="CMIP6",
            experiment_id=["ssp126", "ssp245", "ssp585"],
            variable_id=[
                "hurt",
                "lai",
                "pr",
                "sfcWind",
                "tasmax",
                "tasmin",
                "ts",
                "tos",
            ],
            frequency="day",
            nominal_resolution='"100 km"',
        ),
    ),
]


@pytest.mark.parametrize("facets, expected_query", test_cases)
def test_parse_facets(facets, expected_query):
    result = parse_facets(facets)
    assert result == expected_query



import asyncio
import ssl
from collections.abc import AsyncIterator
from functools import partial
from typing import TypeAlias

from aiostream.stream import merge
from httpx import AsyncClient, HTTPError

from esgpull.config import Config
from esgpull.download import DownloadCtx, Simple
from esgpull.exceptions import DownloadSizeError
from esgpull.fs import Digest, Filesystem
from esgpull.models import File
from esgpull.result import Err, Ok, Result
from esgpull.tui import logger

# Callback: TypeAlias = Callable[[], None] | partial[None]
Callback: TypeAlias = partial[None]

default_ssl_context: ssl.SSLContext | bool = False
default_ssl_context_loaded = False


def load_default_ssl_context() -> str:
    global default_ssl_context
    global default_ssl_context_loaded
    if ssl.OPENSSL_VERSION_INFO[0] >= 3:
        default_ssl_context = ssl.create_default_context()
        default_ssl_context.options |= 0x4
        msg = "Using openssl 3 or higher"
    else:
        default_ssl_context = True
        msg = "Using openssl 1"
    default_ssl_context_loaded = True
    return msg


class Task:
    def __init__(
        self,
        config: Config,
        fs: Filesystem,
        # *,
        # url: str | None = None,
        file: File,
        start_callbacks: list[Callback] | None = None,
    ) -> None:
        self.config = config
        self.fs = fs
        self.ctx = DownloadCtx(file)
        if not self.config.download.disable_checksum:
            self.ctx.digest = Digest(file)
        # if file is None and url is not None:
        #     self.file = self.fetch_file(url)
        # elif file is not None:
        #     self.file = file
        # else:
        #     raise ValueError("no arguments")
        self.downloader = Simple()
        if start_callbacks is None:
            self.start_callbacks = []
        else:
            self.start_callbacks = start_callbacks

    @property
    def file(self) -> File:
        return self.ctx.file

    # def fetch_file(self, url: str) -> File:
    #     ctx = Context()
    #     # [?]TODO: define map data_node->index_node to find url-file
    #     # ctx.query.index_node = ...
    #     ctx.query.title = Path(url).name
    #     results = ctx.search(file=True)
    #     for res in results:
    #         file = File.from_dict(res)
    #         if file.version in url:
    #             return file
    #     raise ValueError(f"{url} is not valid")

    async def stream(
        self,
        semaphore: asyncio.Semaphore,
        client: AsyncClient,
    ) -> AsyncIterator[Result[DownloadCtx]]:
        ctx = self.ctx
        try:
            async with semaphore, self.fs.open(ctx.file) as file_obj:
                for callback in self.start_callbacks:
                    callback()
                stream = self.downloader.stream(
                    client,
                    ctx,
                    self.config.download.chunk_size,
                )
                async for ctx in stream:
                    if ctx.chunk is not None:
                        await file_obj.write(ctx.chunk)
                        ctx.chunk = None
                    if ctx.error:
                        err = DownloadSizeError(ctx.completed, ctx.file.size)
                        yield Err(ctx, err)
                        await stream.aclose()
                        break
                    elif ctx.finished:
                        await file_obj.to_done()
                    yield Ok(ctx)
        except (
            HTTPError,
            DownloadSizeError,
            GeneratorExit,
            ssl.SSLError,
            FileNotFoundError,
            # KeyboardInterrupt,
        ) as err:
            yield Err(ctx, err)


class Processor:
    def __init__(
        self,
        config: Config,
        fs: Filesystem,
        files: list[File],
        start_callbacks: dict[str, list[Callback]],
    ) -> None:
        self.config = config
        self.fs = fs
        self.files = list(filter(self.should_download, files))
        self.tasks: list[Task] = []
        msg: str | None = None
        if not default_ssl_context_loaded:
            msg = load_default_ssl_context()
        self.ssl_context: ssl.SSLContext | bool
        if self.config.download.disable_ssl:
            self.ssl_context = False
        else:
            if msg is not None:
                logger.info(msg)
            self.ssl_context = default_ssl_context
        for file in files:
            task = Task(
                config=config,
                fs=fs,
                file=file,
                start_callbacks=start_callbacks[file.sha],
            )
            self.tasks.append(task)

    def should_download(self, file: File) -> bool:
        if self.fs[file].drs.is_file():
            return False
        else:
            return True

    async def process(self) -> AsyncIterator[Result[DownloadCtx]]:
        semaphore = asyncio.Semaphore(self.config.download.max_concurrent)
        async with AsyncClient(
            follow_redirects=True,
            verify=self.ssl_context,
            timeout=self.config.download.http_timeout,
        ) as client:
            streams = [task.stream(semaphore, client) for task in self.tasks]
            async with merge(*streams).stream() as stream:
                async for result in stream:
                    yield result



import click
from click.exceptions import Abort

from esgpull.cli.decorators import args, opts
from esgpull.cli.utils import init_esgpull
from esgpull.models import sql
from esgpull.tui import Verbosity


@click.command()
@args.key
@opts.verbosity
def facet(
    key: str | None,
    verbosity: Verbosity,
):
    esg = init_esgpull(verbosity)
    with esg.ui.logging("facet", onraise=Abort):
        if key is None:
            results = esg.db.scalars(sql.facet.names())
        else:
            results = esg.db.scalars(sql.facet.values(key))
        esg.ui.print(sorted(results))



# import click
# from click.exceptions import Abort, Exit

# from esgpull import Esgpull
# from esgpull.cli.decorators import opts
# from esgpull.cli.utils import filter_docs, totable
# from esgpull.tui import Verbosity


# @click.command()
# @opts.force
# @opts.verbosity
# def autoremove(
#     force: bool,
#     verbosity: Verbosity,
# ):
#     esg = Esgpull(verbosity=verbosity)
#     with esg.ui.logging("autoremove", onraise=Abort):
#         deprecated = esg.db.get_deprecated_files()
#         nb = len(deprecated)
#         if not nb:
#             esg.ui.print("All files are up to date.")
#             raise Exit(0)
#         if not force:
#             docs = filter_docs([file.raw for file in deprecated])
#             esg.ui.print(totable(docs))
#             s = "s" if nb > 1 else ""
#             esg.ui.print(f"Found {nb} file{s} to remove.")
#             if not esg.ui.ask("Continue?", default=True):
#                 raise Abort
#         removed = esg.remove(*deprecated)
#         esg.ui.print(f"Removed {len(removed)} files with newer version.")
#         nb_remain = len(removed) - nb
#         if nb_remain:
#             esg.ui.print(f"{nb_remain} files could not be removed.")
#         if force:
#             docs = filter_docs([file.raw for file in removed])
#             esg.ui.print(totable(docs))



import pytest
import sqlalchemy as sa

from esgpull import __version__
from esgpull.database import Database
from esgpull.models import Facet, FileStatus, sql


@pytest.fixture
def db(config):
    return Database.from_config(config)


def test_empty(root, db):
    assert str(root) in db.url
    if "+dev" not in __version__:
        assert db.version == __version__


def test_CRUD(db):
    stmt = sa.select(Facet)
    facets = db.scalars(stmt)
    assert len(facets) == 0
    facet = Facet(name="name", value="value")
    facet.compute_sha()
    db.add(facet)
    facets = db.scalars(stmt)
    assert facets == [facet]
    facet.value = "other"
    facet.compute_sha()
    db.add(facet)  # this is an UPDATE as `facet` comes from a SELECT
    facets_with_name = db.scalars(stmt.where(Facet.name == "name"))
    assert len(facets_with_name) == 1
    assert facets_with_name[0].value == "other"
    db.delete(facet)
    facets = db.scalars(stmt)
    assert len(facets) == 0


def test_rows(db):
    facets = [Facet(name=f"name{i}", value=f"value{i}") for i in range(2)]
    for facet in facets:
        facet.compute_sha()
    db.add(*facets)
    stmt = sa.select(Facet)
    rows = db.rows(stmt)
    assert rows == [(facets[0],), (facets[1],)]
    assert db.scalars(stmt.where(Facet.name == "name0")) == [facets[0]]


def test_in(db, file):
    assert file not in db
    db.add(file)
    assert file in db
    db.delete(file)
    assert file not in db


def test_sql_status(db, file):
    db.add(file)
    assert db.scalars(sql.file.with_status(FileStatus.Queued)) == [file]
    assert db.scalars(sql.file.with_status(FileStatus.Done)) == []



import asyncio
import sys

import click
import rich
from click.exceptions import Abort, Exit

if sys.version_info < (3, 11):
    from exceptiongroup import BaseExceptionGroup

from esgpull.cli.decorators import args, opts
from esgpull.cli.utils import get_queries, init_esgpull, valid_name_tag
from esgpull.models import File, FileStatus
from esgpull.tui import Verbosity, logger
from esgpull.utils import format_size


@click.command()
@args.query_id
@opts.tag
@opts.disable_ssl
@opts.quiet
@opts.record
@opts.verbosity
def download(
    query_id: str | None,
    tag: str | None,
    disable_ssl: bool,
    quiet: bool,
    record: bool,
    verbosity: Verbosity,
):
    """
    Asynchronously download files linked to queries
    """
    esg = init_esgpull(verbosity, record=record)
    if disable_ssl:
        esg.config.download.disable_ssl = True
    with esg.ui.logging("download", onraise=Abort):
        if not valid_name_tag(esg.graph, esg.ui, query_id, tag):
            esg.ui.raise_maybe_record(Exit(1))
        if query_id is None and tag is None:
            esg.graph.load_db()
            graph = esg.graph
        else:
            queries = get_queries(esg.graph, query_id, tag)
            graph = esg.graph.subgraph(
                *queries,
                children=True,
                parents=True,
            )
        esg.ui.print(graph)
        shas: set[str] = set()
        queue: list[File] = []
        for query in graph.queries.values():
            for file in query.files:
                if file.status == FileStatus.Queued and file.sha not in shas:
                    shas.add(file.sha)
                    queue.append(file)
        if not queue:
            rich.print("Download queue is empty.")
            esg.ui.raise_maybe_record(Exit(0))
        coro = esg.download(queue, show_progress=not quiet)
        files, errors = asyncio.run(coro)
        if files:
            size = format_size(sum(file.size for file in files))
            esg.ui.print(
                f"Downloaded {len(files)} new files for a total size of {size}"
            )
        if errors:
            logger.error(f"{len(errors)} files could not be installed.")
            exc_group = BaseExceptionGroup("Download", [e.err for e in errors])
            esg.ui.raise_maybe_record(exc_group)
        esg.ui.raise_maybe_record(Exit(0))



# Plugin that raises an exception for testing error isolation

from esgpull.plugin import Event, on

# Track calls for testing
calls = {"file_complete": 0}


@on(
    Event.file_complete, priority="high"
)  # High priority to run before other plugins
def handle_file_complete(file, destination, start_time, end_time, logger):
    """Handler that raises an exception"""
    logger.info(f"About to raise an exception for: {file.filename}")
    calls["file_complete"] += 1
    # Deliberately raise an exception
    raise ValueError(
        f"Deliberate test exception in error_plugin for {file.filename}"
    )



#!/usr/bin/env python3

# https://click.palletsprojects.com/en/latest/
import click

from esgpull import __version__
from esgpull.cli.add import add
from esgpull.cli.config import config
from esgpull.cli.convert import convert
from esgpull.cli.download import download
from esgpull.cli.index_nodes import index_nodes
from esgpull.cli.plugins import plugins
from esgpull.cli.remove import remove
from esgpull.cli.retry import retry
from esgpull.cli.search import search
from esgpull.cli.self import self
from esgpull.cli.show import show
from esgpull.cli.status import status
from esgpull.cli.track import track, untrack
from esgpull.cli.update import update
from esgpull.tui import UI

# from esgpull.cli.autoremove import autoremove
# from esgpull.cli.facet import facet
# from esgpull.cli.get import get
# from esgpull.cli.install import install

# [-]TODO: stats
#   - speed per index/data node
#   - total disk usage
#   - log config for later optimisation ?

SUBCOMMANDS: list[click.Command] = [
    add,
    # autoremove,
    config,
    convert,
    download,
    # facet,
    # get,
    self,
    # install,
    plugins,
    remove,
    retry,
    search,
    show,
    track,
    untrack,
    status,
    # # stats,
    update,
    index_nodes,
]

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])

_ui = UI("/tmp")
version_msg = _ui.render(f"esgpull, version [green]{__version__}[/]")


@click.group(context_settings=CONTEXT_SETTINGS)
@click.version_option(None, "-V", "--version", message=version_msg)
def cli():
    """
    esgpull is a management utility for files and datasets from ESGF.
    """


for subcmd in SUBCOMMANDS:
    cli.add_command(subcmd)


def main():
    cli()
